/*
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "includes.h"

RCSID("$Id: bsd-os2.c,v 1.0 2002/05/23 05:35:34 mouring Exp $");

#ifdef __OS2__

#include "ssh-os2.h"

#define INCL_VIO
#include <os2.h>

#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <io.h>
#include <share.h>
#include <sys/types.h>
#include <dirent.h>
#include <emx/io.h>

/*---------------------------------------------- Directory names replacement -*/

/*
 * These routines are replacement for EMX`s own open(), fopen()
 * taking care of converting such unix-specific things like '/tmp/',
 * '/dev/null' etc into their os/2 equivalents
 */

static struct
{
  char *path;
  int mode;
  char *repl;
  int len;
} conv_path [] =
{
  { "/dev/null", 0, "/dev/nul" },
  { "/tmp/", 1, "TMP" },
  { "/tmp/", 1, "TEMP" },
  { "/etc/", 1, "ETC" },
  { NULL }
};

static char *convert_filename (const char *name, char *dest, int maxlen)
{
  char *s;
  int i;

  for (i = 0; conv_path[i].path; i++)
  {
    if (!conv_path[i].len)
      conv_path[i].len = strlen (conv_path[i].path);
    if (strncmp (name, conv_path[i].path, conv_path[i].len) == 0)
      switch (conv_path[i].mode)
      {
        case 0:
          strncpy (dest, conv_path[i].repl, maxlen);
          return (dest);
        case 1:
          if ((s = getenv (conv_path[i].repl)) && *s)
          {
            strncpy (dest, s, maxlen);
            s = strchr (dest, 0);
            maxlen -= (s - dest);
            if ((s [-1] != '/') && (s [-1] != '\\'))
              *s++ = '/', maxlen--;
            strncpy (s, name + conv_path[i].len, maxlen);
            return (dest);
          }
          break;
      }
  }
  strncpy (dest, name, maxlen);
  return (dest);
}


int binary_pipe(int fd[2])
{
#undef pipe
	int ret = __pipe(fd, 65536);

	if (!ret) {
		setmode (fd[0], O_BINARY);
		setmode (fd[1], O_BINARY);
	}
	return ret;
}

int os2_open (const char *name, int oflag,...)
{
  va_list va;
  int h;
  char *new_name = convert_filename (name, alloca (MAXPATHLEN + 1), MAXPATHLEN + 1);

  if (!new_name)
    return (-1);

  va_start (va, oflag);
  h = _vsopen (new_name, oflag, SH_DENYNO, va);
  va_end (va);
  return h;
}

FILE *os2_fopen (const char *fname, const char *mode)
{
  char *new_fname = convert_filename (fname, alloca (MAXPATHLEN + 1), MAXPATHLEN + 1);

  if (!new_fname)
    return (NULL);

#undef fopen
  return fopen (new_fname, mode);
}

/* These functions are defined here instead of term.c because sshd
   needs to be able to set the size of the console but doesn't need
   a full terminal emulator.
*/

static int last_col = 0, last_row = 0;

int tty_size_changed ()
{
  VIOMODEINFO vi;
  vi.cb = sizeof (vi);
  VioGetMode (&vi, 0);
  return (last_col != vi.col || last_row != vi.row);
}

int tty_setsize (struct winsize *ws)
{
  USHORT cx, cy;
  VIOMODEINFO vi;
  vi.cb = sizeof (vi);
  VioGetMode(&vi, 0);

  VioGetCurPos (&cx, &cy, 0);

  vi.col = ws->ws_col;
  vi.row = ws->ws_row;
  vi.cb = sizeof (vi.cb) + sizeof (vi.fbType) + sizeof (vi.color) +
          sizeof (vi.col) + sizeof (vi.row);
  VioSetMode (&vi, 0);

  VioSetCurPos (cx, cy, 0);
  return 0;
}

int tty_getsize (struct winsize *ws)
{
  VIOMODEINFO vi;
  vi.cb = sizeof (vi);
  VioGetMode (&vi, 0);
  last_col = ws->ws_col = vi.col;
  last_row = ws->ws_row = vi.row;
  ws->ws_xpixel = vi.hres;
  ws->ws_ypixel = vi.vres;
  return 0;
}

int tty_ioctl (int handle, int request, void *arg)
{
  int rc;
//  struct winsize *ws = (struct winsize *)arg;

  if (isatty (handle))
//if(0)
    {
      if (request == TIOCGWINSZ)
        {
//fprintf(stderr,"getsize\n");
          tty_getsize ((struct winsize *)arg);
          return 0;
        }
      else if (request == TIOCSWINSZ)
        {
          tty_setsize ((struct winsize *)arg);
          return 0;
        }
    }
  rc = ioctl (handle, request, arg);
//fprintf(stderr, "%d %d %d %d %d\n", ws->ws_col, ws->ws_row, ws->ws_xpixel, ws->ws_ypixel, rc);
  return rc;
}


/*

passwd.c

Author: Andrew Zabolotny <bit@eltech.ru>

OS/2 routines for password database management

*/

/*
 * $Id$
 * $Log$
 * $Endlog$
 */

#define PASSWORD_DATABASE_FILE	"/etc/passwd"

static struct passwd *database;
static int database_entries = 0;
static void load_database (const char *fname);

#undef getpwuid
struct passwd *os2_getpwuid (uid_t uid)
{
  int i;

  if (!database_entries)
    load_database (PASSWORD_DATABASE_FILE);

  for (i = 0; i < database_entries; i++)
    if (uid == database [i].pw_uid)
      return &database [i];
  return NULL;
}

#undef getpwnam
struct passwd *os2_getpwnam (__const__ char *uname)
{
  int i;

  if (!database_entries)
    load_database (PASSWORD_DATABASE_FILE);

  for (i = 0; i < database_entries; i++)
    if (strcmp (uname, database [i].pw_name) == 0)
      return &database [i];
  return NULL;
}

static void load_database (const char *fname)
{
  FILE *f;
  database_entries = 1;
  database = (struct passwd *)malloc (sizeof (struct passwd));
  database [0] = *getpwuid (0);
  database [0].pw_name = strdup (database [0].pw_name);
  database [0].pw_passwd = strdup (database [0].pw_passwd);
  database [0].pw_comment = strdup (database [0].pw_comment);
  database [0].pw_dir = strdup (database [0].pw_dir);
  database [0].pw_shell = strdup (database [0].pw_shell);

  f = os2_fopen (fname, "rt");
  if (!f) return;
  while (!feof (f))
  {
    char *tmp1, *tmp2;
    struct passwd entry;
    char line [300];
    int eol;
    if (!fgets (line, sizeof (line), f))
      break;

    /* Allocate a new record */
    memset (&entry, 0, sizeof (entry));
    eol = 0;

    /* Fetch username */
    tmp1 = line + strspn (line, " \t");
    if (*tmp1 == '#') continue;
    tmp2 = strpbrk (tmp1, ":\n");
    if (!tmp2) tmp2 = strchr (tmp1, 0), eol = 1; else *tmp2 = 0;
    entry.pw_name = (char *)malloc (tmp2 - tmp1 + 1);
    strcpy (entry.pw_name, tmp1);
    if (eol) goto done;

    /* Fetch password */
    tmp1 = tmp2 + 1;
    tmp2 = strpbrk (tmp1, ":\n");
    if (!tmp2) tmp2 = strchr (tmp1, 0), eol = 1; else *tmp2 = 0;
    entry.pw_passwd = (char *)malloc (tmp2 - tmp1 + 1);
    strcpy (entry.pw_passwd, tmp1);
    if (eol) goto done;

    /* Fetch user id */
    entry.pw_uid = strtol (tmp2 + 1, &tmp1, 10);
    tmp2 = strpbrk (tmp1, ":\n");
    if (!tmp2) goto done;

    /* Fetch group id */
    entry.pw_gid = strtol (tmp2 + 1, &tmp1, 10);
    tmp2 = strpbrk (tmp1, ":\n");
    if (!tmp2) goto done;

    /* Fetch description */
    tmp1 = tmp2 + 1;
    tmp2 = strpbrk (tmp1, ":\n");
    if (!tmp2) tmp2 = strchr (tmp1, 0), eol = 1; else *tmp2 = 0;
    entry.pw_comment = (char *)malloc (tmp2 - tmp1 + 1);
    strcpy (entry.pw_comment, tmp1);
    if (eol) goto done;

    /* Fetch home directory */
    tmp1 = tmp2 + 1;
    tmp2 = strpbrk (tmp1, ":\n");
    if (!tmp2) tmp2 = strchr (tmp1, 0), eol = 1; else *tmp2 = 0;
    entry.pw_dir = (char *)malloc (tmp2 - tmp1 + 1);
    strcpy (entry.pw_dir, tmp1);
    if (eol) goto done;

    /* Fetch shell */
    tmp1 = tmp2 + 1;
    tmp2 = strpbrk (tmp1, ":\n");
    if (!tmp2) tmp2 = strchr (tmp1, 0); else *tmp2 = 0;
    entry.pw_shell = (char *)malloc (tmp2 - tmp1 + 1);
    strcpy (entry.pw_shell, tmp1);

done:
    if (!entry.pw_shell) entry.pw_shell = strdup ("");
    if (!entry.pw_dir) entry.pw_dir = strdup ("");
    if (!entry.pw_comment) entry.pw_comment = strdup ("");
    if (!entry.pw_passwd) entry.pw_passwd = strdup ("");
    if (!entry.pw_name) entry.pw_name = strdup ("");
    if (!entry.pw_gecos) entry.pw_gecos = strdup ("");

    /* Translate "$x" into "x:" in directory/file names */
    if (entry.pw_shell [0] == '$' && entry.pw_shell [1])
      {
        entry.pw_shell [0] = entry.pw_shell [1];
        entry.pw_shell [1] = ':';
      } /* endif */
    if (entry.pw_dir [0] == '$' && entry.pw_dir [1])
      {
        entry.pw_dir [0] = entry.pw_dir [1];
        entry.pw_dir [1] = ':';
      } /* endif */

    /* Look for a database entry with same id or same uid, and replace if found */
    for (eol = 0; eol < database_entries; eol++)
      if (strcmp (database [eol].pw_name, entry.pw_name) == 0
       || database [eol].pw_uid == entry.pw_uid)
        {
          database [eol] = entry;
          eol = -1;
          break;
        } /* endif */

    if (eol >= 0)
      {
        database_entries++;
        database = (struct passwd *)realloc (database,
          database_entries * sizeof (struct passwd));
        database [database_entries - 1] = entry;
      } /* endif */
  }
  fclose (f);
}

#endif /* __OS2__ */